/*
 * This file is part of firk's window manager for x11 (fwm)
 * Copyright (C) 2016-2023 firk <firk@cantconnect.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <fcl/types.h>
#define FCL_SIMPLE_NAMES
#include <fcl/strparse.h>
#include "main.h"
#include "cfgparse.h"
#include "config.h"

#define ac_log 0
#define ac_run 1
#define ac_chdir 2
#define ac_setenv 3

typedef struct {
  unsigned int ln;
  unsigned int type; /* 0=log, 1=shellcmd */
  char const * cmd;
  char const * a2;
} startup_conf;


#define MAX_HOTKEYS 200
#define MAX_WIDGETS 200
#define MAX_STARTUP 200
static hotkey_conf hotkeys_list[MAX_HOTKEYS];
static size_t num_hotkeys, over_hotkeys;
static widget_conf widgets_list[MAX_WIDGETS];
static size_t num_widgets, over_widgets;
static startup_conf startup_list[MAX_STARTUP];
static size_t num_startup, over_startup;
static char const * userconf; /* NOTE: this variable may contain strdup'ed value as well as literal string constant */
       int startup_errors;

       int TASKBAR_WINHEIGHT = DEFAULT_TASKBAR_WINHEIGHT;
#ifndef DEFAULT_TASKBAR_FONT_NAME
#define DEFAULT_TASKBAR_FONT_NAME NULL
#endif
       char const * TASKBAR_FONT_NAME = DEFAULT_TASKBAR_FONT_NAME;
#ifndef DEFAULT_HOLIDAYS_FILE
#define DEFAULT_HOLIDAYS_FILE NULL
#endif
       char const * HOLIDAYS_FILE = DEFAULT_HOLIDAYS_FILE;
       char const * BATTERY_STATUS_FILE;
       char const * BACKLIGHT_FILE;
       int HOOK_BRIGHTNESS_KEYS = 1;
       int WINDOW_SNAP_RANGE = 0;
       int WINDOW_STICK_RANGE = 0;

extern hotkey_conf * get_hotkeys_list(size_t * count) {
  *count = num_hotkeys;
  return hotkeys_list;
}

extern char const * get_hotkey_by_keycode(unsigned int kc, unsigned int shift) {
  unsigned int n;
  hotkey_conf * hk;
  for(hk=hotkeys_list,n=num_hotkeys; n; hk++,n--) if(hk->keycode==kc && hk->shift==shift) return hk->cmd;
  return NULL;
}

extern widget_conf * get_widget_conf_by_name(char const * name) {
  unsigned int n;
  widget_conf * w;
  for(w=widgets_list,n=num_widgets; n; w++,n--) if(!strcmp(w->name,name)) return w;
  return NULL;
}

static void free_hotkey_cmd(char const *cmd) {
  /* TODO: free(cmd) except static strings from set_default_config() and pseudo-cmd constants like OP_CLOSE_WINDOW
     currently we only reading config at startup so it is not very bad to skip this freeing
   */
}

static void add_hotkey(unsigned int keysym, unsigned int shift, char const * cmd) {
  size_t n;
  hotkey_conf * hk;
  for(n=0,hk=hotkeys_list; n<num_hotkeys; n++,hk++) {
    if(hk->keysym==keysym && hk->shift==shift) { free_hotkey_cmd(hk->cmd); hk->cmd = cmd; return; }
  }
  if(n<MAX_HOTKEYS) {
    hk->keysym = keysym;
    hk->shift = shift;
    hk->cmd = cmd;
    num_hotkeys++;
  } else {
    over_hotkeys++;
    free_hotkey_cmd(cmd);
  }
}

static void add_widget(widget_conf *wc) {
  if(num_widgets<MAX_WIDGETS) {
    widgets_list[num_widgets++] = *wc;
  } else {
    over_widgets++;
    free((void*)wc->name);
  }
}

static void add_startup(unsigned int ln, unsigned int type, char const * cmd, char const * a2) {
  if(num_startup<MAX_STARTUP) {
    startup_list[num_startup].ln = ln;
    startup_list[num_startup].type = type;
    startup_list[num_startup].cmd = cmd;
    startup_list[num_startup].a2 = a2;
    num_startup++;
  } else {
    free((void*)cmd);
    if(a2) free((void*)a2);
    over_startup++;
  }
}

static void clear_hotkeys(void) {
  while(num_hotkeys) {
    num_hotkeys--;
    free_hotkey_cmd((void*)(hotkeys_list[num_hotkeys].cmd));
  }
  num_hotkeys = over_hotkeys = 0;
}

static void clear_widgets(void) {
  while(num_widgets) {
    num_widgets--;
    free((void*)(widgets_list[num_widgets].name));
  }
  num_widgets = over_widgets = 0;
}

static void clear_startup(void) {
  while(num_startup) {
    num_startup--;
    if(startup_list[num_startup].cmd) free((void*)startup_list[num_startup].cmd);
    if(startup_list[num_startup].a2) free((void*)startup_list[num_startup].a2);
  }
  num_startup = over_startup = 0;
}

static void set_default_config(void) {
  add_hotkey(XK_Escape, 0, OP_CLOSE_WINDOW);
  add_hotkey(XK_F10,    0, OP_KILL_WINDOW);
  add_hotkey(XK_Tab,    0, OP_NEXT_WINDOW);
  add_hotkey(XK_Tab,    ShiftMask, OP_PREV_WINDOW);
  add_hotkey(XK_Tab,    ControlMask, OP_ZPREV_WINDOW);
  add_hotkey(XK_F,      0, OP_TOGGLE_FULL);
  add_hotkey(XK_grave,  0, OP_TOGGLE_ONTOP);
  add_hotkey(XK_T,      0, "xterm");
  add_hotkey(XK_I,      0, "iceweasel");
  add_hotkey(XK_P,      0, "gnome-screenshot -a");
}

static void parse_hotkey_command(unsigned ln, char const *k, char const *cmd) {
  KeySym ks;
  unsigned int shift;
  char const *arg;
  
  shift = 0;
  if(!strncmp(k,"Ctrl",4) && (k[4]=='-' || k[4]=='+')) { shift|=ControlMask; k+=5; }
  if(!strncmp(k,"Shift",5) && (k[5]=='-' || k[5]=='+')) { shift|=ShiftMask; k+=6; }
  if(!strncmp(k,"Ctrl",4) && (k[4]=='-' || k[4]=='+')) { shift|=ControlMask; k+=5; }
  if((ks = XStringToKeysym(k))==NoSymbol) { logprint("unknown keysym in config line %u", ln); startup_errors = 1; return; }
  if(!strcmp(cmd,"OP_CLOSE_WINDOW")) arg = OP_CLOSE_WINDOW;
  else if(!strcmp(cmd,"OP_KILL_WINDOW")) arg = OP_KILL_WINDOW;
  else if(!strcmp(cmd,"OP_MAXIMIZE_WINDOW")) arg = OP_MAXIMIZE_WINDOW;
  else if(!strcmp(cmd,"OP_MINIMIZE_WINDOW")) arg = OP_MINIMIZE_WINDOW;
  else if(!strcmp(cmd,"OP_TOGGLE_FULL")) arg = OP_TOGGLE_FULL;
  else if(!strcmp(cmd,"OP_NEXT_WINDOW")) arg = OP_NEXT_WINDOW;
  else if(!strcmp(cmd,"OP_PREV_WINDOW")) arg = OP_PREV_WINDOW;
  else if(!strcmp(cmd,"OP_ZPREV_WINDOW")) arg = OP_ZPREV_WINDOW;
  else if(!strcmp(cmd,"OP_TOGGLE_ONTOP")) arg = OP_TOGGLE_ONTOP;
  else if(!strcmp(cmd,"OP_QUIT")) arg = OP_QUIT;
  else if(!strcmp(cmd,"OP_BRIGHTNESS_INC")) arg = OP_BRIGHTNESS_INC;
  else if(!strcmp(cmd,"OP_BRIGHTNESS_DEC")) arg = OP_BRIGHTNESS_DEC;
  else if(!(arg = strdup(cmd))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
  add_hotkey(ks, shift, arg);
  return;
}

static void parse_option_command(unsigned ln, char const *n, char const  *v) {
  if(!strcmp(n,"TASKBAR_WINHEIGHT")) {
    if(str_to_sint(v,&TASKBAR_WINHEIGHT)<0 || TASKBAR_WINHEIGHT<10 || TASKBAR_WINHEIGHT>50) { logprint("invalid option value at config line %u", ln); TASKBAR_WINHEIGHT = DEFAULT_TASKBAR_WINHEIGHT; goto err; }
    return;
  }
  if(!strcmp(n,"TASKBAR_FONT_NAME")) {
    if(!(TASKBAR_FONT_NAME = strdup(v))) { logprint("out of memory while reading config line %u", ln); TASKBAR_FONT_NAME = DEFAULT_TASKBAR_FONT_NAME; goto err; }
    return;
  }
  if(!strcmp(n,"HOLIDAYS_FILE")) {
    if(!(HOLIDAYS_FILE = strdup(v))) { logprint("out of memory while reading config line %u", ln); HOLIDAYS_FILE = DEFAULT_HOLIDAYS_FILE; goto err; }
    return;
  }
  if(!strcmp(n,"BATTERY_STATUS_FILE")) {
    if(!(BATTERY_STATUS_FILE = strdup(v))) { logprint("out of memory while reading config line %u", ln); goto err; }
    return;
  }
  if(!strcmp(n,"BACKLIGHT_FILE")) {
    if(!(BACKLIGHT_FILE = strdup(v))) { logprint("out of memory while reading config line %u", ln); goto err; }
    return;
  }
  if(!strcmp(n,"HOOK_BRIGHTNESS_KEYS")) {
    if(!strcmp(v,"0")) HOOK_BRIGHTNESS_KEYS = 0;
    else if(!strcmp(v,"1")) HOOK_BRIGHTNESS_KEYS = 1;
    else { logprint("invalid option value at config line %u", ln); goto err; }
    return;
  }
  if(!strcmp(n,"WINDOW_SNAP_RANGE")) {
    if(str_to_sint(v,&WINDOW_SNAP_RANGE)<0 || WINDOW_SNAP_RANGE<0 || WINDOW_SNAP_RANGE>100) { logprint("invalid option value at config line %u", ln); WINDOW_SNAP_RANGE = 0; goto err; }
    return;
  }
  if(!strcmp(n,"WINDOW_STICK_RANGE")) {
    if(str_to_sint(v,&WINDOW_STICK_RANGE)<0 || WINDOW_STICK_RANGE<0 || WINDOW_STICK_RANGE>100) { logprint("invalid option value at config line %u", ln); WINDOW_STICK_RANGE = 0; goto err; }
    return;
  }
  logprint("unknown option in config line %u", ln);
err:
  startup_errors = 1;
}

static int parse_wc0(char const **st, widget_zone *base, int *offs) {
  char const *s;
  widget_zone b;
  s = *st;
  if(s=skip_prefix(*st,"screen")) b=w_screen;
  else if(s=skip_prefix(*st,"viewport")) b=w_viewport;
  else if(s=skip_prefix(*st,"taskbar")) b=w_taskbar;
  else { s=*st; b=w_none; }
  if(*s!='+' && *s!='-' && b!=w_none) return -1;
  if(strp_to_sint(&s,offs)<0) return -1;
  *st = s;
  *base = b;
  return 0;
}

static void parse_widget_command(unsigned ln, char const *n, char const *v) {
  widget_conf wc;
  char const *p, *pp;
  unsigned int ap, j;
  /*
   * widget:<widget_name> fixed <x0> <y0> [<w> [<h> [<bw>]]]
   * x0, y0 may be raw pixel numbers, or formula: {screen|viewport}{+-}<value> for coord relative to screen or viewport bottom-right point
   * w, h should be raw numbers, 0=use default
   *
   * widget:<widget_name> float [<zone> [<pref> [<pref> [...]]]] [w=<w>] [h=<h>] [bw=<bw>]
   * zone is {screen|taskbar|viewport} - preferred zone for widget
   * pref is {right|left|bottom|top} - preferred place from most preferred to less preferred
   */
  bzero(&wc, sizeof(wc));
  wc.bw = -1;
  if((p=skip_prefix(v,"fixed")) && (*p==' ' || !*p)) {
    wc.type = w_fixed;
    if(!*p) { logprint("missing <x0> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
    p++;
    if(parse_wc0(&p, &wc.xbase, &wc.xoffs)<0) { logprint("bad <x0> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
    if(!*p) { logprint("missing <y0> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
    if(*(p++)!=' ') { logprint("bad <x0> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
    if(parse_wc0(&p, &wc.ybase, &wc.yoffs)<0) { logprint("bad <y0> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
    if(*p) {
      if(*(p++)!=' ') { logprint("bad <y0> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
      if(strp_to_sint(&p,&wc.w)<0) { logprint("bad <w> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
      if(*p) {
        if(*(p++)!=' ') { logprint("bad <w> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
        if(strp_to_sint(&p,&wc.h)<0) { logprint("bad <h> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
        if(*p) {
          if(*(p++)!=' ') { logprint("bad <h> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
          if(strp_to_sint(&p,&wc.bw)<0) { logprint("bad <bw> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
          if(*p) {
            if(*(p++)!=' ') { logprint("bad <bw> value for widget fixed coords in config line %u", ln); startup_errors = 1; return; }
            logprint("extra arg for widget fixed coords in config line %u", ln); startup_errors = 1; return;
          }
        }
      }
    }
    if(!(wc.name = strdup(n))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_widget(&wc);
    return;
  }
  if((p=skip_prefix(v,"float")) && (*p==' ' || !*p)) {
    wc.type = w_float;
    j = 0;
    if(*p) {
      p++;
      if(pp=skip_prefix(p,"screen")) wc.xbase=w_screen;
      else if(pp=skip_prefix(p,"viewport")) wc.xbase=w_viewport;
      else if(pp=skip_prefix(p,"taskbar")) wc.xbase=w_taskbar;
      else { logprint("bad <zone> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
      p = pp;
      if(*p && *p!=' ') { logprint("bad <zone> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
      while(*p) {
        p++;
        if(!*p) { logprint("bad trailing space for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        else if(pp=skip_prefix(p,"top")) ap = 1;
        else if(pp=skip_prefix(p,"bottom")) ap = 2;
        else if(pp=skip_prefix(p,"left")) ap = 3;
        else if(pp=skip_prefix(p,"right")) ap = 4;
        else break;
        p = pp;
        if(*p && *p!=' ') { logprint("bad <pref> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        if(j>=16) { logprint("too many <pref> values for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        wc.xoffs |= (ap << j);
        j += 3;
      }
      if(*p=='w' && p[1]=='=') {
        p+=2;
        if(strp_to_sint(&p,&wc.w)<0 || wc.w<=0) { logprint("bad <w> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        if(*p) {
          if(*(p++)!=' ') { logprint("bad <w> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
          if(!*p) { logprint("bad trailing space for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        }
      }
      if(*p=='h' && p[1]=='=') {
        p+=2;
        if(strp_to_sint(&p,&wc.h)<0 || wc.h<=0) { logprint("bad <h> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        if(*p) {
          if(*(p++)!=' ') { logprint("bad <h> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
          if(!*p) { logprint("bad trailing space for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        }
      }
      if(*p=='b' && p[1]=='w' && p[2]=='=') {
        p+=3;
        if(strp_to_sint(&p,&wc.bw)<0 || wc.bw<0) { logprint("bad <bw> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        if(*p) {
          if(*(p++)!=' ') { logprint("bad <bw> value for widget float prefs in config line %u", ln); startup_errors = 1; return; }
          if(!*p) { logprint("bad trailing space for widget float prefs in config line %u", ln); startup_errors = 1; return; }
        }
      }
      if(*p) { logprint("extra arg for widget float prefs in config line %u", ln); startup_errors = 1; return; }
    }
    if(!(wc.name = strdup(n))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_widget(&wc);
    return;
  }
  logprint("unknown widget type in config line %u", ln); startup_errors = 1; return;
}

static void parse_command(unsigned ln, char * buf) {
  char c;
  char *b0, *b1, *b2;
  char const *arg, *arg1;
  
  while((c=*buf)==' ' || c=='\t') buf++;
  if(c=='#' || !c) return;
  b0 = buf;
  while((c=*buf) && c!=' ' && c!='\t' && c!='#') buf++;
  *buf++ = 0;
  if(c!=' ' && c!='\t') b1 = "";
  else {
    while((c=*buf)==' ' || c=='\t') buf++;
    b1 = buf;
    while((c=*buf) && c!='#') buf++;
    *buf = 0;
  }
  if(!strcmp(b0,"log")) {
    if(!(arg = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_startup(ln, ac_log, arg, NULL);
    return;
  }
  if(!strcmp(b0,"run")) {
    if(!(arg = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_startup(ln, ac_run, arg, NULL);
    return;
  }
  if(!strcmp(b0,"chdir")) {
    if(!(arg = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_startup(ln, ac_chdir, arg, NULL);
    return;
  }
  if(!strcmp(b0,"setenv")) {
    b2 = b1;
    while((c=*b2) && c!=' ' && c!='\t') b2++;
    if(c) {
      *(b2++) = 0;
      while((c=*b2)==' ' || c=='\t') b2++;
    }
    if(!(arg1 = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    if(!(arg = strdup(b2))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; free((void*)arg1); return; }
    add_startup(ln, ac_setenv, arg, arg1);
    return;
  }
  if(!strcmp(b0,"userconf")) {
    if(userconf) { logprint("duplicate 'userconf' option"); startup_errors = 1; return; }
    if(!(userconf = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    return;
  }
  if(!strncmp(b0,"hotkey:",7)) {
    parse_hotkey_command(ln, b0+7, b1);
    return;
  }
  if(!strncmp(b0,"option:",7)) {
    parse_option_command(ln, b0+7, b1);
    return;
  }
  if(!strncmp(b0,"widget:",7)) {
    parse_widget_command(ln, b0+7, b1);
    return;
  }

  if(!strcmp(b0,"clear-hotkeys")) {
    clear_hotkeys();
    return;
  }
  if(!strcmp(b0,"clear-widgets")) {
    clear_widgets();
    return;
  }
  if(!strcmp(b0,"clear-startup")) {
    clear_startup();
    return;
  }
  logprint("error in config line %u", ln); startup_errors = 1; return;
}

static int parse_config_fd(int fd) {
  char buf[4000], c;
  unsigned int sz, p, p0, ln;
  int r, skip;
  sz = 0;
  r = 1;
  skip = 0;
  ln = 1;
  while(r) {
    if(sz>=sizeof(buf)) { logprint("config line %u too long, exiting", ln); exit(-1); }
    while((r = read(fd, buf+sz, sizeof(buf)-sz))<0) if(errno!=EINTR) return -1;
    for(p=sz,p0=0,sz+=r; p<sz; p++) {
      c = buf[p];
      if(c=='\n') {
        buf[p] = 0;
        if(skip) skip=0; else parse_command(ln, buf+p0);
        p0 = p+1;
        ln++;
      } else if(!skip && (c>=0 && c!=9 && c<32 || c==127)) { logprint("config line %u contains invalid symbols, skipping it", ln); startup_errors = 1; skip = 1; }
    }
    if(p0) if(sz-=p0) bcopy(buf+p0, buf, sz);
  }
  if(sz) { logprint("config line %u (last) non-terminated, skipping it", ln); startup_errors = 1; }
  return 0;
}




static void run_startup_item(startup_conf const * S) {
  char const * s;
  switch(S->type) {
  case ac_log:
    logprint("(config) %s", S->cmd);
    return;
  case ac_run:
    system(S->cmd);
    return;
  case ac_chdir:
    if(!(s = parse_str(S->cmd, 1))) { logprint("error parsing chdir argument at line %u", S->ln); exit(-1); }
    else if(chdir(s)<0) { logprint("chdir(%s) error at line %u", s, S->ln); exit(-1); }
    return;
  case ac_setenv:
    if(!(s = parse_str(S->cmd, 0))) { logprint("error parsing setenv argument at line %u", S->ln); exit(-1); }
    else if(setenv(S->a2, s, 1)<0) { logprint("setenv(%s) error at line %u", S->a2, S->ln); exit(-1); }
    return;
  }
}

static int load_config_file(char const * fn, char const * curname, char const * fbname, int may_missing) {
  int fd;
  logprint("loading config file: %s", fn);
  if(fn[0]=='-' && !fn[1]) {
    if(parse_config_fd(0)<0) { logprint("%sconfig read(%s) i/o error, falling back to %s", curname, fn, fbname); return -1; }
  } else {
    if((fd = open(fn, O_RDONLY|O_NOCTTY))<0) {
      if(errno==ENOENT && may_missing) { logprint("%sconfig file \"%s\" does not exist", curname, fn); return -2; }
      logprint("%sconfig open(%s) failed, falling back to %s", curname, fn, fbname); return -1;
    }
    if(parse_config_fd(fd)<0) { logprint("%sconfig read(%s) i/o error, falling back to %s", curname, fn, fbname); close(fd); return -1; }
    close(fd);
  }
  return 0;
}

#ifndef DEFAULT_CONFIG_FILE
#define DEFAULT_CONFIG_FILE NULL
#endif
#ifndef DEFAULT_USERCONFIG_FILE
#define DEFAULT_USERCONFIG_FILE NULL
#endif

extern void load_configs(unsigned int count, char const * const * fns) {
  unsigned int i;
  int r, ok;
  char const * fn;
  
  if(!count || !(fn=*fns) || !*fn) {
    ok = 0;
    if(DEFAULT_CONFIG_FILE && (r=load_config_file(DEFAULT_CONFIG_FILE,"system-wide ","defaults",1))!=-2) {
      if(r<0) { startup_errors=1; goto fallback; }
      ok = 1;
    }
    if(userconf || (userconf=DEFAULT_USERCONFIG_FILE)) {
      if(!(fn = parse_str(userconf,1))) { logprint("error constructing user-local config-file path"); startup_errors=1; }
      else if(*fn) {
        if(!(fn=strdup(fn))) { logprint("error constructing user-local config-file path: out of memory"); startup_errors=1; }
        else {
          r = load_config_file(fn,"user-local ",ok?"system-wide":"defaults",1);
          free((void*)fn);
          if(r==-2) ;
          else if(r<0) {
            startup_errors = 1;
            if(!ok || !DEFAULT_CONFIG_FILE) goto fallback;
            r = load_config_file(DEFAULT_CONFIG_FILE,"system-wide ","defaults",1);
            if(r<0) goto fallback;
            ok = 1;
          } else ok = 1;
        }
      }
    }
    if(!ok) goto defcfg;
  } else {
    if(load_config_file(fn, "", "defaults", 0)<0) goto fallback;
    for(i=1; i<count; i++) if(load_config_file(fns[i], "secondary ", "primary", 0)<0) {
      clear_hotkeys(); clear_widgets(); clear_startup(); startup_errors = 1;
      if(load_config_file(fn, "", "defaults", 0)<0) goto fallback;
      break;
    }
  }
  if(over_hotkeys) { logprint("too many hotkeys, %u skipped", (unsigned)over_hotkeys); startup_errors = 1; }
  if(over_widgets) { logprint("too many widgets, %u skipped", (unsigned)over_widgets); startup_errors = 1; }
  if(over_startup) { logprint("too many startup commands, %u skipped", (unsigned)over_startup); startup_errors = 1; }
  return;
fallback:
  clear_hotkeys(); clear_widgets(); clear_startup(); startup_errors = 1;
defcfg:
  logprint("using default config"); set_default_config(); return;
}

extern void run_startup(void) {
  unsigned int i;
  for(i=0; i<num_startup; i++) run_startup_item(startup_list+i);
  clear_startup();
}

typedef struct {
  unsigned Y, M, n;
  unsigned char * list;
} t_holidays;

static time_t holidays_list_time;
static t_holidays * holidays_list;
static size_t num_holidays_list;
static int holidays_error;
static t_holidays * holidays_last;
static t_holidays holidays_none;

static void free_holidays_list(t_holidays *l, size_t n) {
  size_t j;
  for(j=0; j<n; j++) free(l[j].list);
  if(n) free(l);
}

static int parse_holiday(unsigned ln, char const * st, t_holidays ** pl, size_t * pn) {
  unsigned Y, M, D, n;
  char c;
  t_holidays * list;
  size_t j, jj;
  unsigned char * ll;

  Y = M = D = 0;
  while(1) {
    c = *(st++);
    if(c<'0' || c>'9') break;
    n = Y*10 + (c-'0');
    if(n<Y) goto err;
    Y = n;
  }
  if(!Y || c!='-') goto err;
  while(1) {
    c = *(st++);
    if(c<'0' || c>'9') break;
    n = M*10 + (c-'0');
    if(n<M) goto err;
    M = n;
  }
  if(M<1 || M>12 || c!='-') goto err;
  while(1) {
    c = *(st++);
    if(c<'0' || c>'9') break;
    n = D*10 + (c-'0');
    if(n<D) goto err;
    D = n;
  }
  if(D<1 || D>31 || c!=' ') goto err;
  c = *(st++);
  if(c<'0' || c>'3' || *st) goto err;
  c = (c=='1');
  list = *pl;
  if(!(j = *pn)) {
    if(!(list = malloc(sizeof(*list)))) goto mem;
    list->Y = Y; list->M = M; list->n = 1;
    if(!(list->list = malloc(2))) { free(list); goto mem; }
    list->list[0] = D; list->list[1] = c;
    *pl = list; *pn = 1;
    return 0;
  }
  if(list[j-1].Y>Y || list[j-1].Y==Y && list[j-1].M>M) {
    logprint("holidays-list line %u not in order, skipping it", ln); return -1;
  }
  if(list[j-1].Y!=Y || list[j-1].M!=M) {
    jj = (j+1)*sizeof(*list);
    if(!jj || jj/sizeof(*list)-1!=j) goto mem;
    if(!(list = realloc(list, jj))) goto mem;
    *pl = list;
    list[j].Y = Y; list[j].M = M; list[j].n = 1;
    if(!(list[j].list = malloc(2))) goto mem;
    list[j].list[0] = D; list[j].list[1] = c;
    *pn = j+1;
    return 0;
  }
  list += j-1;
  if(list->n>=31) {
    logprint("holidays-list line %u - too many entires for smae month, skipping it", ln); return -1;
  }
  if(!(ll = realloc(list->list, (list->n+1)*2))) goto mem;
  list->list = ll;
  ll[list->n*2] = D;
  ll[list->n*2+1] = c;
  list->n++;
  return 0;
err: logprint("holidays-list line %u syntax error, skipping it", ln); return -1;
mem: logprint("out of memory while parsing holidays-list line %u, skipping it", ln); return -1;
}

static int parse_holidays_list_fd(int fd, t_holidays ** pl, size_t * pn) {
  char buf[4000], c;
  unsigned int sz, p, p0, ln;
  int r, skip;

  sz = 0;
  r = 1;
  skip = 0;
  ln = 1;
  while(r) {
    if(sz>=sizeof(buf)) { logprint("holidays-list line %u too long", ln); return -1; }
    while((r = read(fd, buf+sz, sizeof(buf)-sz))<0) if(errno!=EINTR) return -1;
    for(p=sz,p0=0,sz+=r; p<sz; p++) {
      c = buf[p];
      if(c=='\n') {
        buf[p] = 0;
        if(skip) skip=0; else parse_holiday(ln, buf+p0, pl, pn);
        p0 = p+1;
        ln++;
      } else if(!skip && (c>=0 && c!=9 && c<32 || c==127)) { logprint("holidays-list line %u contains invalid symbols, skipping it", ln); skip = 1; }
    }
    if(p0) if(sz-=p0) bcopy(buf+p0, buf, sz);
  }
  if(sz) { logprint("holidays-list line %u (last) non-terminated, skipping it", ln); }
  return 0;
}

extern void check_holidays_file(int force) {
  int fd;
  t_holidays * newlist;
  size_t newn;
  struct stat sb;

  if(!HOLIDAYS_FILE) return;
  while(stat(HOLIDAYS_FILE, &sb)<0) if(errno!=EINTR) {
    logprint("can't stat holidays file, please create it or disable in config: %s", HOLIDAYS_FILE);
    holidays_error = 1;
    return;
  }
  if(!force && sb.st_mtime==holidays_list_time) return;
  if(!holidays_list_time) logprint("loading holidays file: %s", HOLIDAYS_FILE);
  else logprint("reloading holidays file: %s", HOLIDAYS_FILE);
  while((fd = open(HOLIDAYS_FILE, O_RDONLY|O_NOCTTY))<0) if(errno!=EINTR) {
    logprint("can't open holidays file, please create it or disable in config: %s", HOLIDAYS_FILE);
    holidays_error = 1;
    return;
  }
  holidays_list_time = sb.st_mtime;
  newlist = NULL;
  newn = 0;
  if(parse_holidays_list_fd(fd, &newlist, &newn)<0) {
    close(fd);
    free_holidays_list(newlist, newn);
    logprint("error reading or parsing holidays file, please fix it or disable in config: %s", HOLIDAYS_FILE);
    holidays_error = 1;
    return;
  }
  close(fd);
  holidays_last = NULL;
  free_holidays_list(holidays_list, num_holidays_list);
  holidays_list = newlist;
  num_holidays_list = newn;
  holidays_error = 0;
}

static t_holidays * get_holidays_list0(unsigned Y, unsigned M) {
  t_holidays * hl;
  size_t n;
  for(hl=holidays_list,n=0; n<num_holidays_list; n++,hl++) {
    if(hl->Y<Y) continue;
    if(hl->Y==Y) {
      if(hl->M<M) continue;
      if(hl->M==M) return hl;
    }
    break;
  }
  return &holidays_none;
}

extern unsigned char const * get_holidays_list(unsigned Y, unsigned M, size_t * sz) {
  t_holidays * hl;
  hl = get_holidays_list0(Y,M);
  *sz = hl->n;
  return hl->list;
}

extern int get_holiday_state(unsigned Y, unsigned M, unsigned D) {
  unsigned j;
  unsigned char const *l;
  if(!holidays_last || holidays_last->Y!=Y || holidays_last->M!=M) holidays_last = get_holidays_list0(Y,M);
  for(j=holidays_last->n,l=holidays_last->list; j; j--,l+=2) if(l[0]==D) return l[1];
  return -1;
}


